@@ -30,7 +30,6 @@
 #include <gprs_codel.h>
 #include <decoding.h>
 #include <encoding.h>
-#include <gprs_coding_scheme.h>
 #include <gprs_ms.h>
 #include <gprs_ms_storage.h>
 #include <llc.h>
@@ -44,6 +43,7 @@ extern "C" {
 	#include <osmocom/core/linuxlist.h>
 	#include <osmocom/core/logging.h>
 	#include <osmocom/core/rate_ctr.h>
+	#include <osmocom/core/stats.h>
 	#include <osmocom/core/timer.h>
 	#include <osmocom/core/utils.h>
 	#include <osmocom/gsm/gsm_utils.h>
@@ -57,12 +57,42 @@ extern "C" {
 /* After sending these frames, we poll for ack/nack. */
 #define POLL_ACK_AFTER_FRAMES 20
 
-static inline void tbf_update_ms_class(struct gprs_rlcmac_tbf *tbf,
-					const uint8_t ms_class)
-{
-	if (!tbf->ms_class() && ms_class)
-		tbf->set_ms_class(ms_class);
-}
+extern void *tall_pcu_ctx;
+
+static const struct rate_ctr_desc tbf_dl_gprs_ctr_description[] = {
+	{ "gprs:downlink:cs1",              "CS1        " },
+	{ "gprs:downlink:cs2",              "CS2        " },
+	{ "gprs:downlink:cs3",              "CS3        " },
+	{ "gprs:downlink:cs4",              "CS4        " },
+};
+
+static const struct rate_ctr_desc tbf_dl_egprs_ctr_description[] = {
+	{ "egprs:downlink:mcs1",            "MCS1        " },
+	{ "egprs:downlink:mcs2",            "MCS2        " },
+	{ "egprs:downlink:mcs3",            "MCS3        " },
+	{ "egprs:downlink:mcs4",            "MCS4        " },
+	{ "egprs:downlink:mcs5",            "MCS5        " },
+	{ "egprs:downlink:mcs6",            "MCS6        " },
+	{ "egprs:downlink:mcs7",            "MCS7        " },
+	{ "egprs:downlink:mcs8",            "MCS8        " },
+	{ "egprs:downlink:mcs9",            "MCS9        " },
+};
+
+static const struct rate_ctr_group_desc tbf_dl_gprs_ctrg_desc = {
+	"tbf:gprs",
+	"Data Blocks",
+	OSMO_STATS_CLASS_SUBSCRIBER,
+	ARRAY_SIZE(tbf_dl_gprs_ctr_description),
+	tbf_dl_gprs_ctr_description,
+};
+
+static const struct rate_ctr_group_desc tbf_dl_egprs_ctrg_desc = {
+	"tbf:egprs",
+	"Data Blocks",
+	OSMO_STATS_CLASS_SUBSCRIBER,
+	ARRAY_SIZE(tbf_dl_egprs_ctr_description),
+	tbf_dl_egprs_ctr_description,
+};
 
 static void llc_timer_cb(void *_tbf)
 {
@@ -76,8 +106,83 @@ static void llc_timer_cb(void *_tbf)
 	tbf->request_dl_ack();
 }
 
-gprs_rlcmac_dl_tbf::gprs_rlcmac_dl_tbf(BTS *bts_) :
-	gprs_rlcmac_tbf(bts_, GPRS_RLCMAC_DL_TBF),
+gprs_rlcmac_dl_tbf::BandWidth::BandWidth() :
+	dl_bw_octets(0),
+	dl_throughput(0),
+	dl_loss_lost(0),
+	dl_loss_received(0)
+{
+	timespecclear(&dl_bw_tv);
+	timespecclear(&dl_loss_tv);
+}
+
+static int dl_tbf_dtor(struct gprs_rlcmac_dl_tbf *tbf)
+{
+	tbf->~gprs_rlcmac_dl_tbf();
+	return 0;
+}
+
+struct gprs_rlcmac_dl_tbf *tbf_alloc_dl_tbf(struct gprs_rlcmac_bts *bts, GprsMs *ms, int8_t use_trx, bool single_slot)
+{
+	struct gprs_rlcmac_dl_tbf *tbf;
+	int rc;
+
+	OSMO_ASSERT(ms != NULL);
+
+	LOGP(DTBF, LOGL_DEBUG, "********** DL-TBF starts here **********\n");
+	LOGP(DTBF, LOGL_INFO, "Allocating DL TBF: MS_CLASS=%d/%d\n",
+	     ms_ms_class(ms), ms_egprs_ms_class(ms));
+
+	tbf = talloc(tall_pcu_ctx, struct gprs_rlcmac_dl_tbf);
+
+	if (!tbf)
+		return NULL;
+
+	talloc_set_destructor(tbf, dl_tbf_dtor);
+	new (tbf) gprs_rlcmac_dl_tbf(bts, ms);
+
+	rc = tbf->setup(use_trx, single_slot);
+	/* if no resource */
+	if (rc < 0) {
+		talloc_free(tbf);
+		return NULL;
+	}
+
+	if (tbf->is_egprs_enabled()) {
+		tbf->set_window_size();
+		tbf->m_dl_egprs_ctrs = rate_ctr_group_alloc(tbf,
+							&tbf_dl_egprs_ctrg_desc,
+							tbf->m_ctrs->idx);
+		if (!tbf->m_dl_egprs_ctrs) {
+			LOGPTBF(tbf, LOGL_ERROR, "Couldn't allocate EGPRS DL counters\n");
+			talloc_free(tbf);
+			return NULL;
+		}
+	} else {
+		tbf->m_dl_gprs_ctrs = rate_ctr_group_alloc(tbf,
+							&tbf_dl_gprs_ctrg_desc,
+							tbf->m_ctrs->idx);
+		if (!tbf->m_dl_gprs_ctrs) {
+			LOGPTBF(tbf, LOGL_ERROR, "Couldn't allocate GPRS DL counters\n");
+			talloc_free(tbf);
+			return NULL;
+		}
+	}
+
+	llist_add(tbf_bts_list((struct gprs_rlcmac_tbf *)tbf), &bts->dl_tbfs);
+	bts_do_rate_ctr_inc(tbf->bts, CTR_TBF_DL_ALLOCATED);
+
+	tbf->m_last_dl_poll_fn = -1;
+	tbf->m_last_dl_drained_fn = -1;
+
+	osmo_clock_gettime(CLOCK_MONOTONIC, &tbf->m_bw.dl_bw_tv);
+	osmo_clock_gettime(CLOCK_MONOTONIC, &tbf->m_bw.dl_loss_tv);
+
+	return tbf;
+}
+
+gprs_rlcmac_dl_tbf::gprs_rlcmac_dl_tbf(struct gprs_rlcmac_bts *bts_, GprsMs *ms) :
+	gprs_rlcmac_tbf(bts_, ms, GPRS_RLCMAC_DL_TBF),
 	m_tx_counter(0),
 	m_wait_confirm(0),
 	m_dl_ack_requested(false),
@@ -97,28 +202,27 @@ void gprs_rlcmac_dl_tbf::cleanup()
 
 void gprs_rlcmac_dl_tbf::start_llc_timer()
 {
-	if (bts_data()->llc_idle_ack_csec > 0) {
-		struct timeval tv;
-		csecs_to_timeval(bts_data()->llc_idle_ack_csec, &tv);
-		osmo_timer_schedule(&m_llc_timer, tv.tv_sec, tv.tv_usec);
+	if (the_pcu->vty.llc_idle_ack_csec > 0) {
+		struct timespec tv;
+		csecs_to_timespec(the_pcu->vty.llc_idle_ack_csec, &tv);
+		osmo_timer_schedule(&m_llc_timer, tv.tv_sec, tv.tv_nsec / 1000);
 	}
 }
 
-int gprs_rlcmac_dl_tbf::append_data(const uint8_t ms_class,
-				const uint16_t pdu_delay_csec,
-				const uint8_t *data, const uint16_t len)
+int gprs_rlcmac_dl_tbf::append_data(uint16_t pdu_delay_csec,
+				    const uint8_t *data, uint16_t len)
 {
+	struct timespec expire_time;
+
 	LOGPTBFDL(this, LOGL_DEBUG, "appending %u bytes\n", len);
-	gprs_llc_queue::MetaInfo info;
+
 	struct msgb *llc_msg = msgb_alloc(len, "llc_pdu_queue");
 	if (!llc_msg)
 		return -ENOMEM;
 
-	gprs_llc_queue::calc_pdu_lifetime(bts, pdu_delay_csec, &info.expire_time);
-	gettimeofday(&info.recv_time, NULL);
+	gprs_llc_queue::calc_pdu_lifetime(bts, pdu_delay_csec, &expire_time);
 	memcpy(msgb_put(llc_msg, len), data, len);
-	llc_queue()->enqueue(llc_msg, &info);
-	tbf_update_ms_class(this, ms_class);
+	llc_queue()->enqueue(llc_msg, &expire_time);
 	start_llc_timer();
 
 	if (state_is(GPRS_RLCMAC_WAIT_RELEASE)) {
@@ -129,27 +233,15 @@ int gprs_rlcmac_dl_tbf::append_data(const uint8_t ms_class,
 	return 0;
 }
 
-static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts,
-				const char *imsi,
-				const uint32_t tlli, const uint32_t tlli_old,
-				const uint8_t ms_class,
-				const uint8_t egprs_ms_class,
-				struct gprs_rlcmac_dl_tbf **tbf)
+static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts, GprsMs *ms,
+				 struct gprs_rlcmac_dl_tbf **tbf)
 {
 	bool ss;
 	int8_t use_trx;
-	uint16_t ta = GSM48_TA_INVALID;
 	struct gprs_rlcmac_ul_tbf *ul_tbf = NULL, *old_ul_tbf;
 	struct gprs_rlcmac_dl_tbf *dl_tbf = NULL;
-	GprsMs *ms;
 
-	/* check for uplink data, so we copy our informations */
-	ms = bts->bts->ms_store().get_ms(tlli, tlli_old, imsi);
-	if (ms) {
-		ul_tbf = ms->ul_tbf();
-		ta = ms->ta();
-	}
-	/* TODO: if (!ms) create MS before tbf_alloc is called? */
+	ul_tbf = ms_ul_tbf(ms);
 
 	if (ul_tbf && ul_tbf->m_contention_resolution_done
 	 && !ul_tbf->m_final_ack_sent) {
@@ -165,20 +257,15 @@ static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts,
 	// Create new TBF (any TRX)
 /* FIXME: Copy and paste with alloc_ul_tbf */
 	/* set number of downlink slots according to multislot class */
-	dl_tbf = tbf_alloc_dl_tbf(bts, ms, use_trx, ms_class, egprs_ms_class, ss);
+	dl_tbf = tbf_alloc_dl_tbf(bts, ms, use_trx, ss);
 
 	if (!dl_tbf) {
 		LOGP(DTBF, LOGL_NOTICE, "No PDCH resource\n");
 		return -EBUSY;
 	}
-	dl_tbf->update_ms(tlli, GPRS_RLCMAC_DL_TBF);
-	dl_tbf->ms()->set_ta(ta);
 
 	LOGPTBFDL(dl_tbf, LOGL_DEBUG, "[DOWNLINK] START\n");
 
-	/* Store IMSI for later look-up and PCH retransmission */
-	dl_tbf->assign_imsi(imsi);
-
 	/* trigger downlink assignment and set state to ASSIGN.
 	 * we don't use old_downlink, so the possible uplink is used
 	 * to trigger downlink assignment. if there is no uplink,
@@ -191,69 +278,63 @@ static int tbf_new_dl_assignment(struct gprs_rlcmac_bts *bts,
 /**
  * TODO: split into unit test-able parts...
  */
-int gprs_rlcmac_dl_tbf::handle(struct gprs_rlcmac_bts *bts,
-		const uint32_t tlli, const uint32_t tlli_old, const char *imsi,
-		uint8_t ms_class, uint8_t egprs_ms_class,
-		const uint16_t delay_csec,
-		const uint8_t *data, const uint16_t len)
+int dl_tbf_handle(struct gprs_rlcmac_bts *bts,
+		  const uint32_t tlli, const uint32_t tlli_old, const char *imsi,
+		  uint8_t ms_class, uint8_t egprs_ms_class,
+		  const uint16_t delay_csec,
+		  const uint8_t *data, const uint16_t len)
 {
 	struct gprs_rlcmac_dl_tbf *dl_tbf = NULL;
 	int rc;
 	GprsMs *ms, *ms_old;
 
 	/* check for existing TBF */
-	ms = bts->bts->ms_store().get_ms(tlli, tlli_old, imsi);
-	if (ms) {
-		dl_tbf = ms->dl_tbf();
+	ms = bts_ms_store(bts)->get_ms(tlli, tlli_old, imsi);
 
-		/* If we known the GPRS/EGPRS MS class, use it */
-		if (ms->ms_class() || ms->egprs_ms_class()) {
-			ms_class = ms->ms_class();
-			egprs_ms_class = ms->egprs_ms_class();
-		}
-	}
-
-	if (ms && strlen(ms->imsi()) == 0) {
-		ms_old = bts->bts->ms_store().get_ms(0, 0, imsi);
+	if (ms && strlen(ms_imsi(ms)) == 0) {
+		ms_old = bts_ms_store(bts)->get_ms(0, 0, imsi);
 		if (ms_old && ms_old != ms) {
 			/* The TLLI has changed (RAU), so there are two MS
 			 * objects for the same MS */
 			LOGP(DTBF, LOGL_NOTICE,
 			     "There is a new MS object for the same MS: (0x%08x, '%s') -> (0x%08x, '%s')\n",
-			     ms_old->tlli(), ms_old->imsi(), ms->tlli(), ms->imsi());
+			     ms_tlli(ms_old), ms_imsi(ms_old), ms_tlli(ms), ms_imsi(ms));
 
-			GprsMs::Guard guard_old(ms_old);
+			ms_ref(ms_old);
 
-			if (!dl_tbf && ms_old->dl_tbf()) {
+			if (!ms_dl_tbf(ms) && ms_dl_tbf(ms_old)) {
 				LOGP(DTBF, LOGL_NOTICE,
 				     "IMSI %s, old TBF %s: moving DL TBF to new MS object\n",
-				     imsi, ms_old->dl_tbf()->name());
-				dl_tbf = ms_old->dl_tbf();
+				     imsi, ms_dl_tbf(ms_old)->name());
+				dl_tbf = ms_dl_tbf(ms_old);
 				/* Move the DL TBF to the new MS */
 				dl_tbf->set_ms(ms);
 			}
-			/* Clean up the old MS object */
-			/* TODO: Put this into a separate function, use timer? */
-			if (ms_old->ul_tbf() && !ms_old->ul_tbf()->timers_pending(T_MAX))
-				tbf_free(ms_old->ul_tbf());
-			if (ms_old->dl_tbf() && !ms_old->dl_tbf()->timers_pending(T_MAX))
-				tbf_free(ms_old->dl_tbf());
+			ms_merge_and_clear_ms(ms, ms_old);
 
-			ms->merge_old_ms(ms_old);
+			ms_unref(ms_old);
 		}
 	}
 
+	if (!ms)
+		ms = bts_alloc_ms(bts, ms_class, egprs_ms_class);
+	ms_set_imsi(ms, imsi);
+	ms_confirm_tlli(ms, tlli);
+	if (!ms_ms_class(ms) && ms_class) {
+		ms_set_ms_class(ms, ms_class);
+	}
+	if (!ms_egprs_ms_class(ms) && egprs_ms_class) {
+		ms_set_egprs_ms_class(ms, egprs_ms_class);
+	}
+
+	dl_tbf = ms_dl_tbf(ms);
 	if (!dl_tbf) {
-		rc = tbf_new_dl_assignment(bts, imsi, tlli, tlli_old,
-			ms_class, egprs_ms_class, &dl_tbf);
+		rc = tbf_new_dl_assignment(bts, ms, &dl_tbf);
 		if (rc < 0)
 			return rc;
 	}
 
-	/* TODO: ms_class vs. egprs_ms_class is not handled here */
-	rc = dl_tbf->append_data(ms_class, delay_csec, data, len);
-	dl_tbf->update_ms(tlli, GPRS_RLCMAC_DL_TBF);
-	dl_tbf->assign_imsi(imsi);
+	rc = dl_tbf->append_data(delay_csec, data, len);
 
 	return rc;
 }
@@ -261,27 +342,27 @@ int gprs_rlcmac_dl_tbf::handle(struct gprs_rlcmac_bts *bts,
 struct msgb *gprs_rlcmac_dl_tbf::llc_dequeue(bssgp_bvc_ctx *bctx)
 {
 	struct msgb *msg;
-	struct timeval tv_now, tv_now2;
+	struct timespec tv_now, tv_now2;
 	uint32_t octets = 0, frames = 0;
-	struct timeval hyst_delta = {0, 0};
+	struct timespec hyst_delta = {0, 0};
 	const unsigned keep_small_thresh = 60;
-	const gprs_llc_queue::MetaInfo *info;
+	const MetaInfo *info;
 
-	if (bts_data()->llc_discard_csec)
-		csecs_to_timeval(bts_data()->llc_discard_csec, &hyst_delta);
+	if (the_pcu->vty.llc_discard_csec)
+		csecs_to_timespec(the_pcu->vty.llc_discard_csec, &hyst_delta);
 
-	gettimeofday(&tv_now, NULL);
-	timeradd(&tv_now, &hyst_delta, &tv_now2);
+	osmo_clock_gettime(CLOCK_MONOTONIC, &tv_now);
+	timespecadd(&tv_now, &hyst_delta, &tv_now2);
 
 	while ((msg = llc_queue()->dequeue(&info))) {
-		const struct timeval *tv_disc = &info->expire_time;
-		const struct timeval *tv_recv = &info->recv_time;
+		const struct timespec *tv_disc = &info->expire_time;
+		const struct timespec *tv_recv = &info->recv_time;
 
 		gprs_bssgp_update_queue_delay(tv_recv, &tv_now);
 
-		if (ms() && ms()->codel_state()) {
-			int bytes = llc_queue()->octets();
-			if (gprs_codel_control(ms()->codel_state(),
+		if (ms() && ms_codel_state(ms())) {
+			int bytes = llc_queue_octets(llc_queue());
+			if (gprs_codel_control(ms_codel_state(ms()),
 					tv_recv, &tv_now, bytes))
 				goto drop_frame;
 		}
@@ -310,12 +391,12 @@ struct msgb *gprs_rlcmac_dl_tbf::llc_dequeue(bssgp_bvc_ctx *bctx)
 				break;
 		}
 
-		bts->llc_timedout_frame();
+		bts_do_rate_ctr_inc(bts, CTR_LLC_FRAME_TIMEDOUT);
 drop_frame:
 		frames++;
 		octets += msg->len;
 		msgb_free(msg);
-		bts->llc_dropped_frame();
+		bts_do_rate_ctr_inc(bts, CTR_LLC_FRAME_DROPPED);
 		continue;
 	}
 
@@ -323,7 +404,7 @@ drop_frame:
 		LOGPTBFDL(this, LOGL_NOTICE, "Discarding LLC PDU "
 			"because lifetime limit reached, "
 			"count=%u new_queue_size=%zu\n",
-			  frames, llc_queue_size());
+			  frames, llc_queue_size(llc_queue()));
 		if (frames > 0xff)
 			frames = 0xff;
 		if (octets > 0xffffff)
@@ -363,20 +444,28 @@ bool gprs_rlcmac_dl_tbf::restart_bsn_cycle()
 }
 
 int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
-	int previous_bsn, bool *may_combine)
+	int previous_bsn, enum mcs_kind req_mcs_kind, bool *may_combine)
 {
 	int bsn;
 	int data_len2, force_data_len = -1;
-	GprsCodingScheme force_cs;
+	enum CodingScheme tx_cs;
+
+	/* Scheduler may be fine with sending any kind of data, but if
+	   the selected TBF is GPRS-only, then let's filter out EGPRS
+	   here */
+	if (!is_egprs_enabled())
+		req_mcs_kind = GPRS;
 
 	/* search for a nacked or resend marked bsn */
 	bsn = m_window.resend_needed();
 
 	if (previous_bsn >= 0) {
-		force_cs = m_rlc.block(previous_bsn)->cs_current_trans;
-		if (!mcs_is_edge(force_cs))
+		tx_cs = m_rlc.block(previous_bsn)->cs_current_trans;
+		if (!mcs_is_edge(tx_cs))
 			return -1;
 		force_data_len = m_rlc.block(previous_bsn)->len;
+	} else {
+		tx_cs = ms_current_cs_dl(ms(), req_mcs_kind);
 	}
 
 	if (bsn >= 0) {
@@ -390,17 +479,16 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
 
 		if (is_egprs_enabled()) {
 			/* Table 8.1.1.2 and Table 8.1.1.1 of 44.060 */
-			m_rlc.block(bsn)->cs_current_trans = get_retx_mcs(m_rlc.block(bsn)->cs_init,
-									  ms()->current_cs_dl(),
-									  !bts->bts_data()->dl_arq_type);
+			m_rlc.block(bsn)->cs_current_trans = get_retx_mcs(m_rlc.block(bsn)->cs_init, tx_cs,
+									  bts->pcu->vty.dl_arq_type == EGPRS_ARQ1);
 
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "initial_cs_dl(%s) last_mcs(%s) demanded_mcs(%s) cs_trans(%s) arq_type(%d) bsn(%d)\n",
 				  mcs_name(m_rlc.block(bsn)->cs_init),
 				  mcs_name(m_rlc.block(bsn)->cs_last),
-				  mcs_name(ms()->current_cs_dl()),
+				  mcs_name(tx_cs),
 				  mcs_name(m_rlc.block(bsn)->cs_current_trans),
-				  bts->bts_data()->dl_arq_type, bsn);
+				  the_pcu->vty.dl_arq_type, bsn);
 
 			/* TODO: Need to remove this check when MCS-8 -> MCS-6
 			 * transistion is handled.
@@ -421,7 +509,7 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
 		LOGPTBFDL(this, LOGL_DEBUG, "Resending BSN %d\n", bsn);
 		/* re-send block with negative aknowlegement */
 		m_window.m_v_b.mark_unacked(bsn);
-		bts->rlc_resent();
+		bts_do_rate_ctr_inc(bts, CTR_RLC_RESENT);
 	} else if (state_is(GPRS_RLCMAC_FINISHED)) {
 		/* If the TBF is in finished, we already sent all packages at least once.
 		 * If any packages could have been sent (because of unacked) it should have
@@ -429,42 +517,41 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
 		LOGPTBFDL(this, LOGL_DEBUG,
 			  "Restarting at BSN %d, because all blocks have been transmitted.\n",
 			  m_window.v_a());
-		bts->rlc_restarted();
+		bts_do_rate_ctr_inc(bts, CTR_RLC_RESTARTED);
 		if (restart_bsn_cycle())
-			return take_next_bsn(fn, previous_bsn, may_combine);
+			return take_next_bsn(fn, previous_bsn, req_mcs_kind, may_combine);
 	} else if (dl_window_stalled()) {
 		/* There are no more packages to send, but the window is stalled.
 		 * Restart the bsn_cycle to resend all unacked messages */
 		LOGPTBFDL(this, LOGL_NOTICE,
 			  "Restarting at BSN %d, because the window is stalled.\n",
 			  m_window.v_a());
-		bts->rlc_stalled();
+		bts_do_rate_ctr_inc(bts, CTR_RLC_STALLED);
 		if (restart_bsn_cycle())
-			return take_next_bsn(fn, previous_bsn, may_combine);
+			return take_next_bsn(fn, previous_bsn, req_mcs_kind, may_combine);
 	} else if (have_data()) {
 		/* The window has space left, generate new bsn */
-		GprsCodingScheme new_cs;
-		new_cs = force_cs ? force_cs : current_cs();
 		LOGPTBFDL(this, LOGL_DEBUG,
-			  "Sending new block at BSN %d, CS=%s\n",
-			  m_window.v_s(), mcs_name(new_cs));
+			  "Sending new block at BSN %d, CS=%s%s\n",
+			  m_window.v_s(), mcs_name(tx_cs),
+			  force_data_len != -1 ? " (forced)" : "");
 
-		bsn = create_new_bsn(fn, new_cs);
-	} else if (bts->bts_data()->dl_tbf_preemptive_retransmission && !m_window.window_empty()) {
+		bsn = create_new_bsn(fn, tx_cs);
+	} else if (bts->pcu->vty.dl_tbf_preemptive_retransmission && !m_window.window_empty()) {
 		/* The window contains unacked packages, but not acked.
 		 * Mark unacked bsns as RESEND */
 		LOGPTBFDL(this, LOGL_DEBUG,
 			  "Restarting at BSN %d, because all blocks have been transmitted (FLOW).\n",
 			  m_window.v_a());
-		bts->rlc_restarted();
+		bts_do_rate_ctr_inc(bts, CTR_RLC_RESTARTED);
 		if (restart_bsn_cycle())
-			return take_next_bsn(fn, previous_bsn, may_combine);
+			return take_next_bsn(fn, previous_bsn, req_mcs_kind, may_combine);
 	} else {
 		/* Nothing left to send, create dummy LLC commands */
 		LOGPTBFDL(this, LOGL_DEBUG,
 			  "Sending new dummy block at BSN %d, CS=%s\n",
-			  m_window.v_s(), mcs_name(current_cs()));
-		bsn = create_new_bsn(fn, current_cs());
+			  m_window.v_s(), mcs_name(tx_cs));
+		bsn = create_new_bsn(fn, tx_cs);
 		/* Don't send a second block, so don't set cs_current_trans */
 	}
 
@@ -473,11 +560,11 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
 		LOGPTBFDL(this, LOGL_DEBUG,
 			  "Nothing else to send, Re-transmit final block!\n");
 		bsn = m_window.v_s_mod(-1);
-		bts->rlc_final_block_resent();
-		bts->rlc_resent();
+		bts_do_rate_ctr_inc(bts, CTR_RLC_FINAL_BLOCK_RESENT);
+		bts_do_rate_ctr_inc(bts, CTR_RLC_RESENT);
 	}
 
-	*may_combine = num_data_blocks(m_rlc.block(bsn)->cs_current_trans.headerTypeData()) > 1;
+	*may_combine = num_data_blocks(mcs_header_type(m_rlc.block(bsn)->cs_current_trans)) > 1;
 
 	return bsn;
 }
@@ -486,20 +573,20 @@ int gprs_rlcmac_dl_tbf::take_next_bsn(uint32_t fn,
  * Create DL data block
  * The messages are fragmented and forwarded as data blocks.
  */
-struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(uint32_t fn, uint8_t ts)
+struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(uint32_t fn, uint8_t ts, enum mcs_kind req_mcs_kind)
 {
 	int bsn, bsn2 = -1;
 	bool may_combine;
 
-	LOGPTBFDL(this, LOGL_DEBUG, "downlink (V(A)==%d .. V(S)==%d)\n",
-		  m_window.v_a(), m_window.v_s());
+	LOGPTBFDL(this, LOGL_DEBUG, "downlink (V(A)==%d .. V(S)==%d) mcs_mode_restrict=%s\n",
+		  m_window.v_a(), m_window.v_s(), mode_name(req_mcs_kind));
 
-	bsn = take_next_bsn(fn, -1, &may_combine);
+	bsn = take_next_bsn(fn, -1, req_mcs_kind, &may_combine);
 	if (bsn < 0)
 		return NULL;
 
 	if (may_combine)
-		bsn2 = take_next_bsn(fn, bsn, &may_combine);
+		bsn2 = take_next_bsn(fn, bsn, req_mcs_kind, &may_combine);
 
 	return create_dl_acked_block(fn, ts, bsn, bsn2);
 }
@@ -507,6 +594,7 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(uint32_t fn, uint8_t ts)
 /* depending on the current TBF, we assign on PACCH or AGCH */
 void gprs_rlcmac_dl_tbf::trigger_ass(struct gprs_rlcmac_tbf *old_tbf)
 {
+	uint16_t pgroup;
 	/* stop pending timer */
 	stop_timers("assignment (DL-TBF)");
 
@@ -530,7 +618,9 @@ void gprs_rlcmac_dl_tbf::trigger_ass(struct gprs_rlcmac_tbf *old_tbf)
 		TBF_SET_ASS_ON(this, GPRS_RLCMAC_FLAG_CCCH, false);
 
 		/* send immediate assignment */
-		bts->snd_dl_ass(this, false, imsi());
+		if ((pgroup = imsi2paging_group(imsi())) > 999)
+			LOGPTBFDL(this, LOGL_ERROR, "IMSI to paging group failed! (%s)\n", imsi());
+		bts_snd_dl_ass(bts, this, false, pgroup);
 		m_wait_confirm = 1;
 	}
 }
@@ -539,23 +629,23 @@ void gprs_rlcmac_dl_tbf::schedule_next_frame()
 {
 	struct msgb *msg;
 
-	if (m_llc.frame_length() != 0)
+	if (llc_frame_length(&m_llc) != 0)
 		return;
 
 	/* dequeue next LLC frame, if any */
-	msg = llc_dequeue(gprs_bssgp_pcu_current_bctx());
+	msg = llc_dequeue(bts->pcu->bssgp.bctx);
 	if (!msg)
 		return;
 
 	LOGPTBFDL(this, LOGL_DEBUG, "Dequeue next LLC (len=%d)\n", msg->len);
 
 	m_llc.put_frame(msg->data, msg->len);
-	bts->llc_frame_sched();
+	bts_do_rate_ctr_inc(bts, CTR_LLC_FRAME_SCHED);
 	msgb_free(msg);
 	m_last_dl_drained_fn = -1;
 }
 
-int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, GprsCodingScheme cs)
+int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, enum CodingScheme cs)
 {
 	uint8_t *data;
 	gprs_rlc_data *rlc_data;
@@ -565,17 +655,17 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, GprsCodingScheme cs)
 	int write_offset = 0;
 	Encoding::AppendResult ar;
 
-	if (m_llc.frame_length() == 0)
+	if (llc_frame_length(&m_llc) == 0)
 		schedule_next_frame();
 
-	OSMO_ASSERT(cs.isValid());
+	OSMO_ASSERT(mcs_is_valid(cs));
 
 	/* length of usable data block (single data unit w/o header) */
-	const uint8_t block_data_len = cs.maxDataBlockBytes();
+	const uint8_t block_data_len = mcs_max_data_block_bytes(cs);
 
 	/* now we still have untransmitted LLC data, so we fill mac block */
 	rlc_data = m_rlc.block(bsn);
-	data = rlc_data->prepare(block_data_len);
+	data = prepare(rlc_data, block_data_len);
 	rlc_data->cs_last = cs;
 	rlc_data->cs_current_trans = cs;
 
@@ -597,10 +687,34 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, GprsCodingScheme cs)
 		bool is_final;
 		int payload_written = 0;
 
-		if (m_llc.frame_length() == 0) {
-			/* nothing to sent - delay the release of the TBF */
+		if (llc_frame_length(&m_llc) == 0) {
+			/* It is not clear, when the next real data will
+			 * arrive, so request a DL ack/nack now */
+			request_dl_ack();
 
 			int space = block_data_len - write_offset;
+
+			if (num_chunks != 0) {
+				/* Nothing to send, and we already put some data in
+				 * rlcmac data block, we are done */
+				LOGPTBFDL(this, LOGL_DEBUG,
+					  "LLC queue completely drained and there's "
+					  "still %d free bytes in rlcmac data block\n", space);
+				if (mcs_is_edge(cs)) {
+					/* in EGPRS there's no M bit, so we need
+					 * to flag padding with LI=127 */
+					Encoding::rlc_data_to_dl_append_egprs_li_padding(rdbi,
+											 &write_offset,
+											 &num_chunks,
+											 data);
+				}
+				break;
+			}
+
+			/* Nothing to send from upper layers (LLC), but still
+			 * requested to send something to MS to delay the
+			 * release of the TBF. See 3GPP TS 44.060 9.3.1a
+			 * "Delayed release of downlink Temporary Block Flow" */
 			/* A header will need to by added, so we just need
 			 * space-1 octets */
 			m_llc.put_dummy_frame(space - 1);
@@ -609,29 +723,25 @@ int gprs_rlcmac_dl_tbf::create_new_bsn(const uint32_t fn, GprsCodingScheme cs)
 			if (m_last_dl_drained_fn < 0)
 				m_last_dl_drained_fn = fn;
 
-			/* It is not clear, when the next real data will
-			 * arrive, so request a DL ack/nack now */
-			request_dl_ack();
-
 			LOGPTBFDL(this, LOGL_DEBUG,
 				  "Empty chunk, added LLC dummy command of size %d, drained_since=%d\n",
-				  m_llc.frame_length(), frames_since_last_drain(fn));
+				  llc_frame_length(&m_llc), frames_since_last_drain(fn));
 		}
 
-		is_final = llc_queue_size() == 0 && !keep_open(fn);
+		is_final = llc_queue_size(llc_queue()) == 0 && !keep_open(fn);
 
 		ar = Encoding::rlc_data_to_dl_append(rdbi, cs,
 			&m_llc, &write_offset, &num_chunks, data, is_final, &payload_written);
 
 		if (payload_written > 0)
-			bts->rlc_dl_payload_bytes(payload_written);
+			bts_do_rate_ctr_add(bts, CTR_RLC_DL_PAYLOAD_BYTES, payload_written);
 
 		if (ar == Encoding::AR_NEED_MORE_BLOCKS)
 			break;
 
-		LOGPTBFDL(this, LOGL_DEBUG, "Complete DL frame, len=%d\n", m_llc.frame_length());
-		gprs_rlcmac_dl_bw(this, m_llc.frame_length());
-		bts->llc_dl_bytes(m_llc.frame_length());
+		LOGPTBFDL(this, LOGL_DEBUG, "Complete DL frame, len=%d\n", llc_frame_length(&m_llc));
+		gprs_rlcmac_dl_bw(this, llc_frame_length(&m_llc));
+		bts_do_rate_ctr_add(bts, CTR_LLC_DL_BYTES, llc_frame_length(&m_llc));
 		m_llc.reset();
 
 		if (is_final) {
@@ -685,7 +795,7 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 	int rc;
 	bool is_final = false;
 	gprs_rlc_data_info rlc;
-	GprsCodingScheme cs;
+	enum CodingScheme cs;
 	int bsns[ARRAY_SIZE(rlc.block_info)];
 	unsigned num_bsns;
 	bool need_padding = false;
@@ -734,8 +844,8 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 		 * Refer commit be881c028fc4da00c4046ecd9296727975c206a3
 		 * dated 2016-02-07 23:45:40 (UTC)
 		 */
-		if (cs != GprsCodingScheme(MCS8))
-			cs.decToSingleBlock(&need_padding);
+		if (cs != MCS8)
+			mcs_dec_to_single_block(&cs, &need_padding);
 	}
 
 	spb = get_egprs_dl_spb(index);
@@ -750,7 +860,7 @@ struct msgb *gprs_rlcmac_dl_tbf::create_dl_acked_block(
 	rlc.tfi = m_tfi; /* TFI */
 
 	/* return data block(s) as message */
-	msg_len = cs.sizeDL();
+	msg_len = mcs_size_dl(cs);
 	dl_msg = msgb_alloc(msg_len, "rlcmac_dl_data");
 	if (!dl_msg)
 		return NULL;
@@ -987,7 +1097,7 @@ int gprs_rlcmac_dl_tbf::analyse_errors(char *show_rbb, uint8_t ssn,
 	return lost * 100 / (lost + received);
 }
 
-gprs_rlc_dl_window *gprs_rlcmac_dl_tbf::window()
+gprs_rlc_window *gprs_rlcmac_dl_tbf::window()
 {
 	return &m_window;
 }
@@ -1014,8 +1124,8 @@ int gprs_rlcmac_dl_tbf::update_window(unsigned first_bsn,
 
 	error_rate = analyse_errors(show_rbb, behind_last_bsn, &ana_res);
 
-	if (bts_data()->cs_adj_enabled && ms())
-		ms()->update_error_rate(this, error_rate);
+	if (the_pcu->vty.cs_adj_enabled && ms())
+		ms_update_error_rate(ms(), this, error_rate);
 
 	m_window.update(bts, rbb, first_bsn, &lost, &received);
 	rate_ctr_add(&m_ctrs->ctr[TBF_CTR_RLC_NACKED], lost);
@@ -1069,8 +1179,8 @@ int gprs_rlcmac_dl_tbf::update_window(const uint8_t ssn, const uint8_t *rbb)
 
 	error_rate = analyse_errors(show_rbb, ssn, &ana_res);
 
-	if (bts_data()->cs_adj_enabled && ms())
-		ms()->update_error_rate(this, error_rate);
+	if (the_pcu->vty.cs_adj_enabled && ms())
+		ms_update_error_rate(ms(), this, error_rate);
 
 	m_window.update(bts, show_rbb, ssn,
 			&lost, &received);
@@ -1105,7 +1215,7 @@ int gprs_rlcmac_dl_tbf::maybe_start_new_window()
 	release();
 
 	/* check for LLC PDU in the LLC Queue */
-	if (llc_queue_size() > 0)
+	if (llc_queue_size(llc_queue()) > 0)
 		/* we have more data so we will re-use this tbf */
 		establish_dl_tbf_on_pacch();
 
@@ -1215,8 +1325,8 @@ bool gprs_rlcmac_dl_tbf::need_control_ts() const
 
 bool gprs_rlcmac_dl_tbf::have_data() const
 {
-	return m_llc.chunk_size() > 0 ||
-		(llc_queue_size() > 0);
+	return llc_chunk_size(&m_llc) > 0 ||
+		(llc_queue_size(llc_queue()) > 0);
 }
 
 static inline int frames_since_last(int32_t last, unsigned fn)
@@ -1247,7 +1357,7 @@ bool gprs_rlcmac_dl_tbf::keep_open(unsigned fn) const
 	int keep_time_frames;
 	unsigned long dl_tbf_idle_msec;
 
-	dl_tbf_idle_msec = osmo_tdef_get(bts_data()->T_defs_pcu, -2031, OSMO_TDEF_MS, -1);
+	dl_tbf_idle_msec = osmo_tdef_get(the_pcu->T_defs, -2031, OSMO_TDEF_MS, -1);
 	if (dl_tbf_idle_msec == 0)
 		return false;
 
@@ -1267,11 +1377,11 @@ enum egprs_rlc_dl_reseg_bsn_state
 	egprs_rlc_dl_reseg_bsn_state *block_status_dl =
 				&rlc_data->spb_status.block_status_dl;
 
-	enum CodingScheme cs_init = CodingScheme(rlc_data->cs_init);
-	enum CodingScheme cs_current_trans = CodingScheme(rlc_data->cs_current_trans);
+	enum CodingScheme cs_init = rlc_data->cs_init;
+	enum CodingScheme cs_current_trans = rlc_data->cs_current_trans;
 
-	enum HeaderType ht_cs_init = rlc_data->cs_init.headerTypeData();
-	enum HeaderType ht_cs_current_trans = rlc_data->cs_current_trans.headerTypeData();
+	enum HeaderType ht_cs_init = mcs_header_type(rlc_data->cs_init);
+	enum HeaderType ht_cs_current_trans = mcs_header_type(rlc_data->cs_current_trans);
 
 	*block_data = &rlc_data->block[0];
 
@@ -1340,11 +1450,11 @@ enum egprs_rlcmac_dl_spb gprs_rlcmac_dl_tbf::get_egprs_dl_spb(const int bsn)
 	struct gprs_rlc_data *rlc_data = m_rlc.block(bsn);
 	egprs_rlc_dl_reseg_bsn_state block_status_dl = rlc_data->spb_status.block_status_dl;
 
-	enum CodingScheme cs_init = CodingScheme(rlc_data->cs_init);
-	enum CodingScheme cs_current_trans = CodingScheme(rlc_data->cs_current_trans);
+	enum CodingScheme cs_init = rlc_data->cs_init;
+	enum CodingScheme cs_current_trans = rlc_data->cs_current_trans;
 
-	enum HeaderType ht_cs_init = rlc_data->cs_init.headerTypeData();
-	enum HeaderType ht_cs_current_trans = rlc_data->cs_current_trans.headerTypeData();
+	enum HeaderType ht_cs_init = mcs_header_type(rlc_data->cs_init);
+	enum HeaderType ht_cs_current_trans = mcs_header_type(rlc_data->cs_current_trans);
 
 	/* Table 10.4.8b.1 of 44.060 */
 	if (ht_cs_current_trans == HEADER_EGPRS_DATA_TYPE_3) {
@@ -1354,15 +1464,15 @@ enum egprs_rlcmac_dl_spb gprs_rlcmac_dl_tbf::get_egprs_dl_spb(const int bsn)
 		 */
 		if (block_status_dl == EGPRS_RESEG_FIRST_SEG_SENT) {
 			/* statistics */
-			bts->spb_downlink_second_segment();
+			bts_do_rate_ctr_inc(bts, CTR_SPB_DL_SECOND_SEGMENT);
 			return EGPRS_RLCMAC_DL_SEC_SEG;
 		} else if ((ht_cs_init == HEADER_EGPRS_DATA_TYPE_1) ||
 			   (ht_cs_init == HEADER_EGPRS_DATA_TYPE_2)) {
-			bts->spb_downlink_first_segment();
+			bts_do_rate_ctr_inc(bts, CTR_SPB_DL_FIRST_SEGMENT);
 			return EGPRS_RLCMAC_DL_FIRST_SEG;
 		} else if ((cs_init == MCS4) &&
 			   (cs_current_trans == MCS1)) {
-			bts->spb_downlink_first_segment();
+			bts_do_rate_ctr_inc(bts, CTR_SPB_DL_FIRST_SEGMENT);
 			return EGPRS_RLCMAC_DL_FIRST_SEG;
 		}
 	}
@@ -1372,11 +1482,11 @@ enum egprs_rlcmac_dl_spb gprs_rlcmac_dl_tbf::get_egprs_dl_spb(const int bsn)
 
 void gprs_rlcmac_dl_tbf::set_window_size()
 {
-	const struct gprs_rlcmac_bts *b = bts->bts_data();
+	const struct gprs_rlcmac_bts *b = bts;
 	uint16_t ws = egprs_window_size(b, dl_slots());
 
 	LOGPTBFDL(this, LOGL_INFO, "setting EGPRS DL window size to %u, base(%u) slots(%u) ws_pdch(%u)\n",
-		  ws, b->ws_base, pcu_bitcount(dl_slots()), b->ws_pdch);
+		  ws, bts->pcu->vty.ws_base, pcu_bitcount(dl_slots()), bts->pcu->vty.ws_pdch);
 	m_window.set_ws(ws);
 }
 
@@ -1384,55 +1494,55 @@ void gprs_rlcmac_dl_tbf::update_coding_scheme_counter_dl(enum CodingScheme cs)
 {
 	switch (cs) {
 	case CS1:
-		bts->gprs_dl_cs1();
+		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS1);
 		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS1]);
 		break;
 	case CS2:
-		bts->gprs_dl_cs2();
+		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS2);
 		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS2]);
 		break;
 	case CS3:
-		bts->gprs_dl_cs3();
+		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS3);
 		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS3]);
 		break;
 	case CS4:
-		bts->gprs_dl_cs4();
+		bts_do_rate_ctr_inc(bts, CTR_GPRS_DL_CS4);
 		rate_ctr_inc(&m_dl_gprs_ctrs->ctr[TBF_CTR_GPRS_DL_CS4]);
 		break;
 	case MCS1:
-		bts->egprs_dl_mcs1();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS1);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS1]);
 		break;
 	case MCS2:
-		bts->egprs_dl_mcs2();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS2);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS2]);
 		break;
 	case MCS3:
-		bts->egprs_dl_mcs3();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS3);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS3]);
 		break;
 	case MCS4:
-		bts->egprs_dl_mcs4();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS4);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS4]);
 		break;
 	case MCS5:
-		bts->egprs_dl_mcs5();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS5);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS5]);
 		break;
 	case MCS6:
-		bts->egprs_dl_mcs6();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS6);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS6]);
 		break;
 	case MCS7:
-		bts->egprs_dl_mcs7();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS7);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS7]);
 		break;
 	case MCS8:
-		bts->egprs_dl_mcs8();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS8);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS8]);
 		break;
 	case MCS9:
-		bts->egprs_dl_mcs9();
+		bts_do_rate_ctr_inc(bts, CTR_EGPRS_DL_MCS9);
 		rate_ctr_inc(&m_dl_egprs_ctrs->ctr[TBF_CTR_EGPRS_DL_MCS9]);
 		break;
 	default:
@@ -1440,3 +1550,11 @@ void gprs_rlcmac_dl_tbf::update_coding_scheme_counter_dl(enum CodingScheme cs)
 			  mcs_name(cs));
 	}
 }
+
+struct gprs_rlcmac_dl_tbf *as_dl_tbf(struct gprs_rlcmac_tbf *tbf)
+{
+	if (tbf && tbf->direction == GPRS_RLCMAC_DL_TBF)
+		return static_cast<gprs_rlcmac_dl_tbf *>(tbf);
+	else
+		return NULL;
+}
